# coding: utf-8
from m3 import ApplicationLogicException
import objectpack
import xlrd

from .XLSReader import XLSLoader, START_ROW
from .proxy import fabricate_proxies
from .proxy_import import ProxyLoader
from .ui import BaseImportWindow


class BaseImportPack(objectpack.BasePack):
    """Базовый пак, реализующий действия по импорту"""

    title = u''
    import_window = BaseImportWindow
    # TODO окно показа лога импорта
    # result_window = BaseImportResultWindow

    # постраничные загрузчики
    loaders = {}
    # разрешенные расширения файлов
    extensions = []

    # Необходимо ли для получения loader'ов иметь возможность прочитать
    # данные файла. Ecли True, то в параметрах функции get_loaders придет
    # аргумент book, иначе пердварительного чтения не будет
    workbook_pre_reading = False

    def __init__(self):
        super(BaseImportPack, self).__init__()
        self.import_window_action = BaseImportWindowAction()
        self.import_action = BaseImportAction()
        self.actions.extend([
            self.import_window_action,
            self.import_action,
        ])

    def _make_proxies_config(self, memory_file, file_name, initial_context):
        """Автогенерация прокси у загрузчиков для областей ячеек"""
        xls_loader = XLSLoader(memory_file)
        # только страницы, для которых определены загрузчики
        loaders = dict(  # имена сделаю большими и без пробелов
            map(
                lambda (k, v): (unicode(k).strip().upper(), v),
                self.loaders.iteritems()
            )
        )
        sheets = filter(
            lambda sheet: sheet.name.strip().upper() in loaders.keys(),
            xls_loader._book.sheets()
        )
        for sheet in sheets:
            loader = loaders.get(sheet.name.strip().upper())
            # только если загрузчик имеет прокси, определяющие некую область
            if not hasattr(loader, 'layout_proxies_template'):
                continue

            add_log = {}  # лог ошибок, связанных с добавлением прокси

            # избавление от сгенерированных прокси предыдущего листа
            # при многолистовом импорте разных областей
            loader.proxies = filter(
                lambda proxy: not getattr(proxy[1], 'was_fabricated', False),
                loader.proxies
            )
            # разбор шапок колонок, начиная со строки заголовков START_ROW
            start_row = loader.xlsreader_extra.get(START_ROW, 0)
            head_data = [cell_val for cell_val in sheet.row_values(start_row)]
            for layout_str, ancestor_cls in getattr(
                loader, 'layout_proxies_template', []
            ):
                # сгенерированные прокси для этой области
                proxies = fabricate_proxies(
                    ancestor_cls, layout_str, head_data
                )
                # дополнение загрузчика сгенерированными прокси
                for (key, proxy) in proxies:
                    if hasattr(loader, 'add_proxy'):
                        loader.add_proxy(key, proxy, add_log, initial_context)
        # будет еще одно чтение
        memory_file.seek(0)

    def get_loaders(self, request, context, book=None, **kwargs):
        """
        Получение постраничных загрузчиков
        """
        assert self.loaders, u'Не определены прокси загрузки!'
        loaders = self.loaders.copy()
        return loaders

    def set_initial_context(self, request, context):
        """
        Метод позволяет изменять начальный контекст
        прокси загрузчика в зависимости от контекста,
        """
        initial_context = {}
        return initial_context

    def make_import(self, request, context):
        """Метод осуществляет импорт. Возвращает (лог, True/False)"""

        _file = request.FILES.get('file_uploaded')
        file_name = request.REQUEST.get('uploaded')
        initial_context = self.set_initial_context(request, context)

        _kwargs = {
            'initial_context': initial_context,
        }
        if self.workbook_pre_reading:
            book = xlrd.open_workbook(file_contents=_file.read())
            _file.seek(0)
            _kwargs['book'] = book

        loaders = self.get_loaders(request, context, **_kwargs)

        # если загрузчик предполагает наличие прокси для областей ячеек
        # которые заранее невозможно задекларировать в описании
        self._make_proxies_config(_file, file_name, initial_context)
        # Загрузчик
        loader = ProxyLoader(loaders, _file, file_name, initial_context)

        res = loader.load()
        return loader.message, res

    def get_import_window_params(self, params, request, context):
        """Параметры передаваемые окну импорта"""
        params['extensions'] = self.extensions
        return params

    def extend_menu(self, menu):
        """Размещение в меню"""
        return menu.SubMenu(
            u'Администрирование',
            menu.SubMenu(
                u'Импорт',
                menu.Item(self.title, self.import_window_action)
            )
        )


class BaseImportWindowAction(objectpack.BaseWindowAction):
    """Экшн показа окна импорта"""

    perm_code = 'import'

    def create_window(self):
        self.win = self.parent.import_window()

    def set_window_params(self):
        super(BaseImportWindowAction, self).set_window_params()
        params = self.win_params.copy()
        params['title'] = self.parent.title
        params['form_url'] = self.parent.import_action.get_absolute_url()
        self.win_params = self.parent.get_import_window_params(
            params, self.request, self.context
        )

    def configure_window(self):
        pass


class BaseImportAction(objectpack.BaseAction):
    """Экшн выполняющий импорт"""

    perm_code = 'import'

    def run(self, request, context):
        try:
            log_msg, success = self.parent.make_import(request, context)
        except xlrd.XLRDError:
            raise ApplicationLogicException(
                u'Файл имеет неверный формат или поврежден! '
                u'Пересохраните файл в формате '
                u'"Microsoft Excel 97/2003 (.xls)"'
            )

        return objectpack.multiline_text_window_result(
            success=success, data=log_msg, title=self.parent.title)
